@@ -13,41 +13,155 @@ import asyncio
 import contextlib
 import errno
 import logging
+import socket
 import struct
+from abc import ABC, abstractmethod
 from pprint import pformat as pf
 from typing import Dict, Generator, Optional, Union
 
 # When support for cpython older than 3.11 is dropped
 # async_timeout can be replaced with asyncio.timeout
 from async_timeout import timeout as asyncio_timeout
+from cryptography.hazmat.primitives import hashes
 
+from .deviceconfig import DeviceConfig
 from .exceptions import SmartDeviceException
 from .json import dumps as json_dumps
 from .json import loads as json_loads
 
 _LOGGER = logging.getLogger(__name__)
 _NO_RETRY_ERRORS = {errno.EHOSTDOWN, errno.EHOSTUNREACH, errno.ECONNREFUSED}
+_UNSIGNED_INT_NETWORK_ORDER = struct.Struct(">I")
 
 
-class TPLinkSmartHomeProtocol:
+def md5(payload: bytes) -> bytes:
+    """Return an md5 hash of the payload."""
+    digest = hashes.Hash(hashes.MD5())  # noqa: S303
+    digest.update(payload)
+    hash = digest.finalize()
+    return hash
+
+
+class BaseTransport(ABC):
+    """Base class for all TP-Link protocol transports."""
+
+    DEFAULT_TIMEOUT = 5
+
+    def __init__(
+        self,
+        *,
+        config: DeviceConfig,
+    ) -> None:
+        """Create a protocol object."""
+        self._config = config
+        self._host = config.host
+        self._port = config.port_override or self.default_port
+        self._credentials = config.credentials
+        self._credentials_hash = config.credentials_hash
+        self._timeout = config.timeout
+
+    @property
+    @abstractmethod
+    def default_port(self) -> int:
+        """The default port for the transport."""
+
+    @property
+    @abstractmethod
+    def credentials_hash(self) -> str:
+        """The hashed credentials used by the transport."""
+
+    @abstractmethod
+    async def send(self, request: str) -> Dict:
+        """Send a message to the device and return a response."""
+
+    @abstractmethod
+    async def close(self) -> None:
+        """Close the transport.  Abstract method to be overriden."""
+
+
+class TPLinkProtocol(ABC):
+    """Base class for all TP-Link Smart Home communication."""
+
+    def __init__(
+        self,
+        *,
+        transport: BaseTransport,
+    ) -> None:
+        """Create a protocol object."""
+        self._transport = transport
+
+    @property
+    def _host(self):
+        return self._transport._host
+
+    @property
+    def config(self) -> DeviceConfig:
+        """Return the connection parameters the device is using."""
+        return self._transport._config
+
+    @abstractmethod
+    async def query(self, request: Union[str, Dict], retry_count: int = 3) -> Dict:
+        """Query the device for the protocol.  Abstract method to be overriden."""
+
+    @abstractmethod
+    async def close(self) -> None:
+        """Close the protocol.  Abstract method to be overriden."""
+
+
+class _XorTransport(BaseTransport):
+    """Implementation of the Xor encryption transport.
+
+    WIP, currently only to ensure consistent __init__ method signatures
+    for protocol classes.  Will eventually incorporate the logic from
+    TPLinkSmartHomeProtocol to simplify the API and re-use the IotProtocol
+    class.
+    """
+
+    DEFAULT_PORT: int = 9999
+
+    def __init__(self, *, config: DeviceConfig) -> None:
+        super().__init__(config=config)
+
+    @property
+    def default_port(self):
+        """Default port for the transport."""
+        return self.DEFAULT_PORT
+
+    @property
+    def credentials_hash(self) -> str:
+        """The hashed credentials used by the transport."""
+        return ""
+
+    async def send(self, request: str) -> Dict:
+        """Send a message to the device and return a response."""
+        return {}
+
+    async def close(self) -> None:
+        """Close the transport.  Abstract method to be overriden."""
+
+
+class TPLinkSmartHomeProtocol(TPLinkProtocol):
     """Implementation of the TP-Link Smart Home protocol."""
 
     INITIALIZATION_VECTOR = 171
     DEFAULT_PORT = 9999
-    DEFAULT_TIMEOUT = 5
     BLOCK_SIZE = 4
 
     def __init__(
-        self, host: str, *, port: Optional[int] = None, timeout: Optional[int] = None
+        self,
+        *,
+        transport: BaseTransport,
     ) -> None:
         """Create a protocol object."""
-        self.host = host
-        self.port = port or TPLinkSmartHomeProtocol.DEFAULT_PORT
+        super().__init__(transport=transport)
+
         self.reader: Optional[asyncio.StreamReader] = None
         self.writer: Optional[asyncio.StreamWriter] = None
         self.query_lock = asyncio.Lock()
         self.loop: Optional[asyncio.AbstractEventLoop] = None
-        self.timeout = timeout or TPLinkSmartHomeProtocol.DEFAULT_TIMEOUT
+
+        self._timeout = self._transport._timeout
+        self._port = self._transport._port
 
     async def query(self, request: Union[str, Dict], retry_count: int = 3) -> Dict:
         """Request information from a TP-Link SmartHome Device.
@@ -63,7 +177,7 @@ class TPLinkSmartHomeProtocol:
             assert isinstance(request, str)  # noqa: S101
 
         async with self.query_lock:
-            return await self._query(request, retry_count, self.timeout)
+            return await self._query(request, retry_count, self._timeout)  # type: ignore[arg-type]
 
     async def _connect(self, timeout: int) -> None:
         """Try to connect or reconnect to the device."""
@@ -71,40 +185,51 @@ class TPLinkSmartHomeProtocol:
             return
         self.reader = self.writer = None
 
-        task = asyncio.open_connection(self.host, self.port)
+        task = asyncio.open_connection(self._host, self._port)
         async with asyncio_timeout(timeout):
             self.reader, self.writer = await task
+            sock: socket.socket = self.writer.get_extra_info("socket")
+            # Ensure our packets get sent without delay as we do all
+            # our writes in a single go and we do not want any buffering
+            # which would needlessly delay the request or risk overloading
+            # the buffer on the device
+            sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
 
     async def _execute_query(self, request: str) -> Dict:
         """Execute a query on the device and wait for the response."""
         assert self.writer is not None  # noqa: S101
         assert self.reader is not None  # noqa: S101
         debug_log = _LOGGER.isEnabledFor(logging.DEBUG)
-
         if debug_log:
-            _LOGGER.debug("%s >> %s", self.host, request)
+            _LOGGER.debug("%s >> %s", self._host, request)
         self.writer.write(TPLinkSmartHomeProtocol.encrypt(request))
         await self.writer.drain()
 
         packed_block_size = await self.reader.readexactly(self.BLOCK_SIZE)
-        length = struct.unpack(">I", packed_block_size)[0]
+        length = _UNSIGNED_INT_NETWORK_ORDER.unpack(packed_block_size)[0]
 
         buffer = await self.reader.readexactly(length)
         response = TPLinkSmartHomeProtocol.decrypt(buffer)
         json_payload = json_loads(response)
         if debug_log:
-            _LOGGER.debug("%s << %s", self.host, pf(json_payload))
+            _LOGGER.debug("%s << %s", self._host, pf(json_payload))
 
         return json_payload
 
     async def close(self) -> None:
         """Close the connection."""
         writer = self.writer
+        self.close_without_wait()
+        if writer:
+            with contextlib.suppress(Exception):
+                await writer.wait_closed()
+
+    def close_without_wait(self) -> None:
+        """Close the connection without waiting for the connection to close."""
+        writer = self.writer
         self.reader = self.writer = None
         if writer:
             writer.close()
-            with contextlib.suppress(Exception):
-                await writer.wait_closed()
 
     def _reset(self) -> None:
         """Clear any varibles that should not survive between loops."""
@@ -127,25 +252,35 @@ class TPLinkSmartHomeProtocol:
             except ConnectionRefusedError as ex:
                 await self.close()
                 raise SmartDeviceException(
-                    f"Unable to connect to the device: {self.host}:{self.port}: {ex}"
+                    f"Unable to connect to the device: {self._host}:{self._port}: {ex}"
                 ) from ex
             except OSError as ex:
                 await self.close()
                 if ex.errno in _NO_RETRY_ERRORS or retry >= retry_count:
                     raise SmartDeviceException(
                         f"Unable to connect to the device:"
-                        f" {self.host}:{self.port}: {ex}"
+                        f" {self._host}:{self._port}: {ex}"
                     ) from ex
                 continue
             except Exception as ex:
                 await self.close()
                 if retry >= retry_count:
-                    _LOGGER.debug("Giving up on %s after %s retries", self.host, retry)
+                    _LOGGER.debug("Giving up on %s after %s retries", self._host, retry)
                     raise SmartDeviceException(
                         f"Unable to connect to the device:"
-                        f" {self.host}:{self.port}: {ex}"
+                        f" {self._host}:{self._port}: {ex}"
                     ) from ex
                 continue
+            except BaseException as ex:
+                # Likely something cancelled the task so we need to close the connection
+                # as we are not in an indeterminate state
+                self.close_without_wait()
+                _LOGGER.debug(
+                    "%s: BaseException during connect, closing connection: %s",
+                    self._host,
+                    ex,
+                )
+                raise
 
             try:
                 assert self.reader is not None  # noqa: S101
@@ -155,14 +290,24 @@ class TPLinkSmartHomeProtocol:
             except Exception as ex:
                 await self.close()
                 if retry >= retry_count:
-                    _LOGGER.debug("Giving up on %s after %s retries", self.host, retry)
+                    _LOGGER.debug("Giving up on %s after %s retries", self._host, retry)
                     raise SmartDeviceException(
-                        f"Unable to query the device {self.host}:{self.port}: {ex}"
+                        f"Unable to query the device {self._host}:{self._port}: {ex}"
                     ) from ex
 
                 _LOGGER.debug(
-                    "Unable to query the device %s, retrying: %s", self.host, ex
+                    "Unable to query the device %s, retrying: %s", self._host, ex
                 )
+            except BaseException as ex:
+                # Likely something cancelled the task so we need to close the connection
+                # as we are not in an indeterminate state
+                self.close_without_wait()
+                _LOGGER.debug(
+                    "%s: BaseException during query, closing connection: %s",
+                    self._host,
+                    ex,
+                )
+                raise
 
         # make mypy happy, this should never be reached..
         await self.close()
@@ -192,7 +337,7 @@ class TPLinkSmartHomeProtocol:
         :return: ciphertext to be send over wire, in bytes
         """
         plainbytes = request.encode()
-        return struct.pack(">I", len(plainbytes)) + bytes(
+        return _UNSIGNED_INT_NETWORK_ORDER.pack(len(plainbytes)) + bytes(
             TPLinkSmartHomeProtocol._xor_payload(plainbytes)
         )
 
