@@ -17,29 +17,19 @@ import inspect
 import logging
 from dataclasses import dataclass
 from datetime import datetime, timedelta
-from enum import Enum, auto
 from typing import Any, Dict, List, Optional, Set
 
+from .credentials import Credentials
+from .device_type import DeviceType
+from .deviceconfig import DeviceConfig
 from .emeterstatus import EmeterStatus
 from .exceptions import SmartDeviceException
 from .modules import Emeter, Module
-from .protocol import TPLinkSmartHomeProtocol
+from .protocol import TPLinkProtocol, TPLinkSmartHomeProtocol, _XorTransport
 
 _LOGGER = logging.getLogger(__name__)
 
 
-class DeviceType(Enum):
-    """Device type enum."""
-
-    Plug = auto()
-    Bulb = auto()
-    Strip = auto()
-    StripSocket = auto()
-    Dimmer = auto()
-    LightStrip = auto()
-    Unknown = -1
-
-
 @dataclass
 class WifiNetwork:
     """Wifi network container."""
@@ -52,6 +42,9 @@ class WifiNetwork:
     channel: Optional[int] = None
     rssi: Optional[int] = None
 
+    # For SMART devices
+    signal_level: Optional[int] = None
+
 
 def merge(d, u):
     """Update dict recursively."""
@@ -70,7 +63,7 @@ def requires_update(f):
         @functools.wraps(f)
         async def wrapped(*args, **kwargs):
             self = args[0]
-            if self._last_update is None:
+            if self._last_update is None and f.__name__ not in self._sys_info:
                 raise SmartDeviceException(
                     "You need to await update() to access the data"
                 )
@@ -81,7 +74,7 @@ def requires_update(f):
         @functools.wraps(f)
         def wrapped(*args, **kwargs):
             self = args[0]
-            if self._last_update is None:
+            if self._last_update is None and f.__name__ not in self._sys_info:
                 raise SmartDeviceException(
                     "You need to await update() to access the data"
                 )
@@ -91,11 +84,17 @@ def requires_update(f):
     return wrapped
 
 
+@functools.lru_cache
+def _parse_features(features: str) -> Set[str]:
+    """Parse features string."""
+    return set(features.split(":"))
+
+
 class SmartDevice:
     """Base class for all supported device types.
 
-    You don't usually want to construct this class which implements the shared common interfaces.
-    The recommended way is to either use the discovery functionality, or construct one of the subclasses:
+    You don't usually want to initialize this class manually,
+    but either use :class:`Discover` class, or use one of the subclasses:
 
     * :class:`SmartPlug`
     * :class:`SmartBulb`
@@ -138,7 +137,8 @@ class SmartDevice:
         >>> dev.mac
         01:23:45:67:89:ab
 
-        When initialized using discovery or using a subclass, you can check the type of the device:
+        When initialized using discovery or using a subclass,
+        you can check the type of the device:
 
         >>> dev.is_bulb
         False
@@ -147,7 +147,8 @@ class SmartDevice:
         >>> dev.is_plug
         True
 
-        You can also get the hardware and software as a dict, or access the full device response:
+        You can also get the hardware and software as a dict,
+        or access the full device response:
 
         >>> dev.hw_info
         {'sw_ver': '1.2.5 Build 171213 Rel.101523',
@@ -168,7 +169,8 @@ class SmartDevice:
         >>> dev.is_on
         True
 
-        Some devices provide energy consumption meter, and regular update will already fetch some information:
+        Some devices provide energy consumption meter,
+        and regular update will already fetch some information:
 
         >>> dev.has_emeter
         True
@@ -177,7 +179,8 @@ class SmartDevice:
         >>> dev.emeter_today
         >>> dev.emeter_this_month
 
-        You can also query the historical data (note that these needs to be awaited), keyed with month/day:
+        You can also query the historical data (note that these needs to be awaited),
+        keyed with month/day:
 
         >>> asyncio.run(dev.get_emeter_monthly(year=2016))
         {11: 1.089, 12: 1.582}
@@ -188,33 +191,71 @@ class SmartDevice:
 
     emeter_type = "emeter"
 
-    def __init__(self, host: str) -> None:
+    def __init__(
+        self,
+        host: str,
+        *,
+        config: Optional[DeviceConfig] = None,
+        protocol: Optional[TPLinkProtocol] = None,
+    ) -> None:
         """Create a new SmartDevice instance.
 
         :param str host: host name or ip address on which the device listens
         """
-        self.host = host
-
-        self.protocol = TPLinkSmartHomeProtocol(host)
+        if config and protocol:
+            protocol._transport._config = config
+        self.protocol: TPLinkProtocol = protocol or TPLinkSmartHomeProtocol(
+            transport=_XorTransport(config=config or DeviceConfig(host=host)),
+        )
         _LOGGER.debug("Initializing %s of type %s", self.host, type(self))
         self._device_type = DeviceType.Unknown
-        # TODO: typing Any is just as using Optional[Dict] would require separate checks in
-        #       accessors. the @updated_required decorator does not ensure mypy that these
-        #       are not accessed incorrectly.
+        # TODO: typing Any is just as using Optional[Dict] would require separate
+        #       checks in accessors. the @updated_required decorator does not ensure
+        #       mypy that these are not accessed incorrectly.
         self._last_update: Any = None
+        self._discovery_info: Optional[Dict[str, Any]] = None
+
         self._sys_info: Any = None  # TODO: this is here to avoid changing tests
+        self._features: Set[str] = set()
         self.modules: Dict[str, Any] = {}
 
         self.children: List["SmartDevice"] = []
 
+    @property
+    def host(self) -> str:
+        """The device host."""
+        return self.protocol._transport._host
+
+    @host.setter
+    def host(self, value):
+        """Set the device host.
+
+        Generally used by discovery to set the hostname after ip discovery.
+        """
+        self.protocol._transport._host = value
+        self.protocol._transport._config.host = value
+
+    @property
+    def port(self) -> int:
+        """The device port."""
+        return self.protocol._transport._port
+
+    @property
+    def credentials(self) -> Optional[Credentials]:
+        """The device credentials."""
+        return self.protocol._transport._credentials
+
+    @property
+    def credentials_hash(self) -> Optional[str]:
+        """The protocol specific hash of the credentials the device is using."""
+        return self.protocol._transport.credentials_hash
+
     def add_module(self, name: str, module: Module):
         """Register a module."""
         if name in self.modules:
             _LOGGER.debug("Module %s already registered, ignoring..." % name)
             return
 
-        assert name not in self.modules
-
         _LOGGER.debug("Adding module %s", module)
         self.modules[name] = module
 
@@ -274,11 +315,7 @@ class SmartDevice:
     @requires_update
     def features(self) -> Set[str]:
         """Return a set of features that the device supports."""
-        try:
-            return set(self.sys_info["feature"].split(":"))
-        except KeyError:
-            _LOGGER.debug("Device does not have feature information")
-            return set()
+        return self._features
 
     @property  # type: ignore
     @requires_update
@@ -311,11 +348,12 @@ class SmartDevice:
         # See #105, #120, #161
         if self._last_update is None:
             _LOGGER.debug("Performing the initial update to obtain sysinfo")
-            self._last_update = await self.protocol.query(req)
-            self._sys_info = self._last_update["system"]["get_sysinfo"]
+            response = await self.protocol.query(req)
+            self._last_update = response
+            self._set_sys_info(response["system"]["get_sysinfo"])
 
         await self._modular_update(req)
-        self._sys_info = self._last_update["system"]["get_sysinfo"]
+        self._set_sys_info(self._last_update["system"]["get_sysinfo"])
 
     async def _modular_update(self, req: dict) -> None:
         """Execute an update query."""
@@ -325,40 +363,89 @@ class SmartDevice:
             )
             self.add_module("emeter", Emeter(self, self.emeter_type))
 
+        request_list = []
+        est_response_size = 1024 if "system" in req else 0
         for module in self.modules.values():
             if not module.is_supported:
                 _LOGGER.debug("Module %s not supported, skipping" % module)
                 continue
+
+            est_response_size += module.estimated_query_response_size
+            if est_response_size > self.max_device_response_size:
+                request_list.append(req)
+                req = {}
+                est_response_size = module.estimated_query_response_size
+
             q = module.query()
             _LOGGER.debug("Adding query for %s: %s", module, q)
             req = merge(req, q)
+        request_list.append(req)
 
-        self._last_update = await self.protocol.query(req)
+        responses = [
+            await self.protocol.query(request) for request in request_list if request
+        ]
 
-    def update_from_discover_info(self, info):
+        # Preserve the last update and merge
+        # responses on top of it so we remember
+        # which modules are not supported, otherwise
+        # every other update will query for them
+        update: Dict = self._last_update.copy() if self._last_update else {}
+        for response in responses:
+            update = {**update, **response}
+        self._last_update = update
+
+    def update_from_discover_info(self, info: Dict[str, Any]) -> None:
         """Update state from info from the discover call."""
-        self._last_update = info
-        self._sys_info = info["system"]["get_sysinfo"]
+        self._discovery_info = info
+        if "system" in info and (sys_info := info["system"].get("get_sysinfo")):
+            self._last_update = info
+            self._set_sys_info(sys_info)
+        else:
+            # This allows setting of some info properties directly
+            # from partial discovery info that will then be found
+            # by the requires_update decorator
+            self._set_sys_info(info)
+
+    def _set_sys_info(self, sys_info: Dict[str, Any]) -> None:
+        """Set sys_info."""
+        self._sys_info = sys_info
+        if features := sys_info.get("feature"):
+            self._features = _parse_features(features)
+        else:
+            self._features = set()
 
     @property  # type: ignore
     @requires_update
     def sys_info(self) -> Dict[str, Any]:
-        """Return system information."""
+        """
+        Return system information.
+
+        Do not call this function from within the SmartDevice
+        class itself as @requires_update will be affected for other properties.
+        """
         return self._sys_info  # type: ignore
 
     @property  # type: ignore
     @requires_update
     def model(self) -> str:
         """Return device model."""
-        sys_info = self.sys_info
+        sys_info = self._sys_info
         return str(sys_info["model"])
 
+    @property
+    def has_children(self) -> bool:
+        """Return true if the device has children devices."""
+        # Ideally we would check for the 'child_num' key in sys_info,
+        # but devices that speak klap do not populate this key via
+        # update_from_discover_info so we check for the devices
+        # we know have children instead.
+        return self.is_strip
+
     @property  # type: ignore
-    @requires_update
-    def alias(self) -> str:
+    def alias(self) -> Optional[str]:
         """Return device name (alias)."""
-        sys_info = self.sys_info
-        return str(sys_info["alias"])
+        sys_info = self._sys_info
+        return sys_info.get("alias") if sys_info else None
 
     async def set_alias(self, alias: str) -> None:
         """Set the device name (alias)."""
@@ -409,14 +496,14 @@ class SmartDevice:
             "oemId",
             "dev_name",
         ]
-        sys_info = self.sys_info
+        sys_info = self._sys_info
         return {key: sys_info[key] for key in keys if key in sys_info}
 
     @property  # type: ignore
     @requires_update
     def location(self) -> Dict:
         """Return geographical location."""
-        sys_info = self.sys_info
+        sys_info = self._sys_info
         loc = {"latitude": None, "longitude": None}
 
         if "latitude" in sys_info and "longitude" in sys_info:
@@ -434,7 +521,7 @@ class SmartDevice:
     @requires_update
     def rssi(self) -> Optional[int]:
         """Return WiFi signal strength (rssi)."""
-        rssi = self.sys_info.get("rssi")
+        rssi = self._sys_info.get("rssi")
         return None if rssi is None else int(rssi)
 
     @property  # type: ignore
@@ -444,14 +531,14 @@ class SmartDevice:
 
         :return: mac address in hexadecimal with colons, e.g. 01:23:45:67:89:ab
         """
-        sys_info = self.sys_info
-
+        sys_info = self._sys_info
         mac = sys_info.get("mac", sys_info.get("mic_mac"))
         if not mac:
             raise SmartDeviceException(
                 "Unknown mac, please submit a bug report with sys_info output."
             )
-
+        mac = mac.replace("-", ":")
+        # Format a mac that has no colons (usually from mic_mac field)
         if ":" not in mac:
             mac = ":".join(format(s, "02x") for s in bytes.fromhex(mac))
 
@@ -562,13 +649,13 @@ class SmartDevice:
     @requires_update
     def on_since(self) -> Optional[datetime]:
         """Return pretty-printed on-time, or None if not available."""
-        if "on_time" not in self.sys_info:
+        if "on_time" not in self._sys_info:
             return None
 
         if self.is_off:
             return None
 
-        on_time = self.sys_info["on_time"]
+        on_time = self._sys_info["on_time"]
 
         return datetime.now().replace(microsecond=0) - timedelta(seconds=on_time)
 
@@ -607,7 +694,7 @@ class SmartDevice:
 
         return [WifiNetwork(**x) for x in info["ap_list"]]
 
-    async def wifi_join(self, ssid, password, keytype=3):  # noqa: D202
+    async def wifi_join(self, ssid: str, password: str, keytype: str = "3"):  # noqa: D202
         """Join the given wifi network.
 
         If joining the network fails, the device will return to AP mode after a while.
@@ -616,7 +703,7 @@ class SmartDevice:
         async def _join(target, payload):
             return await self._query_helper(target, "set_stainfo", payload)
 
-        payload = {"ssid": ssid, "password": password, "key_type": keytype}
+        payload = {"ssid": ssid, "password": password, "key_type": int(keytype)}
         try:
             return await _join("netif", payload)
         except SmartDeviceException as ex:
@@ -641,6 +728,11 @@ class SmartDevice:
             )
         return self.children[index]
 
+    @property
+    def max_device_response_size(self) -> int:
+        """Returns the maximum response size the device can safely construct."""
+        return 16 * 1024
+
     @property
     def device_type(self) -> DeviceType:
         """Return the device type."""
@@ -698,9 +790,44 @@ class SmartDevice:
         The returned object contains the raw results from the last update call.
         This should only be used for debugging purposes.
         """
-        return self._last_update
+        return self._last_update or self._discovery_info
 
     def __repr__(self):
         if self._last_update is None:
             return f"<{self._device_type} at {self.host} - update() needed>"
-        return f"<{self._device_type} model {self.model} at {self.host} ({self.alias}), is_on: {self.is_on} - dev specific: {self.state_information}>"
+        return (
+            f"<{self._device_type} model {self.model} at {self.host}"
+            f" ({self.alias}), is_on: {self.is_on}"
+            f" - dev specific: {self.state_information}>"
+        )
+
+    @property
+    def config(self) -> DeviceConfig:
+        """Return the device configuration."""
+        return self.protocol.config
+
+    @staticmethod
+    async def connect(
+        *,
+        host: Optional[str] = None,
+        config: Optional[DeviceConfig] = None,
+    ) -> "SmartDevice":
+        """Connect to a single device by the given hostname or device configuration.
+
+        This method avoids the UDP based discovery process and
+        will connect directly to the device.
+
+        It is generally preferred to avoid :func:`discover_single()` and
+        use this function instead as it should perform better when
+        the WiFi network is congested or the device is not responding
+        to discovery requests.
+
+        :param host: Hostname of device to query
+        :param config: Connection parameters to ensure the correct protocol
+            and connection options are used.
+        :rtype: SmartDevice
+        :return: Object for querying/controlling found device.
+        """
+        from .device_factory import connect  # pylint: disable=import-outside-toplevel
+
+        return await connect(host=host, config=config)  # type: ignore[arg-type]
